@@ -24,33 +24,32 @@ import __main__
 
 import sys
 import os
-import string
-import urllib
+from urllib.parse import quote as url_quote
+from urllib.parse import unquote as url_unquote
 import signal
 import subprocess
 import traceback
 import errno
 import time
-import re
 import pipes
 import socket
 import shutil
 
-debuglevel = None
+import adtlog
+
 progname = "<VirtSubproc>"
-devnull_read = file('/dev/null', 'r')
+devnull_read = open('/dev/null', 'r')
 caller = __main__
 copy_timeout = int(os.getenv('ADT_VIRT_COPY_TIMEOUT', '300'))
 
-downtmp = None
-down = None
-downkind = None
-downs = None
+downtmp_open = None  # downtmp after opening testbed
+downtmp = None  # current downtmp (None after close)
+auxverb = None  # prefix to run command argv in testbed
 cleaning = False
 in_mainloop = False
 
 
-class Quit:
+class Quit(RuntimeError):
 
     def __init__(self, ec, m):
         self.ec = ec
@@ -74,18 +73,12 @@ def timeout_stop():
     signal.alarm(0)
 
 
-class FailedCmd:
+class FailedCmd(RuntimeError):
 
     def __init__(self, e):
         self.e = e
 
 
-def debug(m):
-    if not debuglevel:
-        return
-    print >> sys.stderr, progname + ": debug:", m
-
-
 def bomb(m):
     if in_mainloop:
         raise Quit(12, progname + ": failure: %s" % m)
@@ -96,7 +89,7 @@ def bomb(m):
 
 
 def ok():
-    print 'ok'
+    print('ok')
 
 
 def cmdnumargs(c, ce, nargs=0, noptargs=0):
@@ -108,7 +101,7 @@ def cmdnumargs(c, ce, nargs=0, noptargs=0):
 
 def cmd_capabilities(c, ce):
     cmdnumargs(c, ce)
-    return caller.hook_capabilities() + ['execute-debug']
+    return caller.hook_capabilities()
 
 
 def cmd_quit(c, ce):
@@ -123,67 +116,74 @@ def cmd_close(c, ce):
     cleanup()
 
 
-def cmd_print_auxverb_command(c, ce):
-    return print_command('auxverb', c, ce)
+def cmd_print_execute_command(c, ce):
+    global auxverb
 
-
-def cmd_print_shstring_command(c, ce):
-    return print_command('shstring', c, ce)
-
-
-def print_command(which, c, ce):
-    global downs
     cmdnumargs(c, ce)
     if not downtmp:
-        bomb("`print-%s-command' when not open" % which)
-    cl = downs[which]
-    if not len(cl):
-        cl = ['sh', '-c', 'exec "$@"', 'x'] + cl
-    return [','.join(map(urllib.quote, cl))]
+        bomb("`print-execute-command' when not open")
+    return [','.join(map(url_quote, auxverb))]
 
 
 def preexecfn():
     caller.hook_forked_inchild()
 
 
-def execute_raw(what, instr, timeout, *popenargs, **popenargsk):
-    debug(" ++ %s" % string.join(popenargs[0]))
-    sp = subprocess.Popen(preexec_fn=preexecfn, *popenargs, **popenargsk)
+def execute_timeout(instr, timeout, *popenargs, **popenargsk):
+    '''Popen wrapper with timeout supervision
+
+    If instr is given, it is fed into stdin, otherwise stdin will be /dev/null.
+
+    Return (status, stdout, stderr)
+    '''
+    adtlog.debug('execute-timeout: ' + ' '.join(popenargs[0]))
+    sp = subprocess.Popen(*popenargs,
+                          preexec_fn=preexecfn,
+                          universal_newlines=True,
+                          **popenargsk)
     if instr is None:
         popenargsk['stdin'] = devnull_read
     timeout_start(timeout)
-    (out, err) = sp.communicate(instr)
+    try:
+        (out, err) = sp.communicate(instr)
+    except Timeout:
+        sp.kill()
+        sp.wait()
+        raise
     timeout_stop()
-    if err:
-        bomb("%s unexpectedly produced stderr output `%s'" %
-            (what, err))
     status = sp.wait()
-    return (status, out)
+    return (status, out, err)
 
 
-def execute(cmd_string, cmd_list=[], downp=False, outp=False, timeout=0):
-    cmdl = cmd_string.split()
+def check_exec(argv, downp=False, outp=False, timeout=0):
+    '''Run successful command (argv list)
+
+    Command must succeed (exit code 0) and not produce any stderr. If downp is
+    True, command is run in testbed. If outp is True, stdout will be captured
+    and returned. stdin is set to /dev/null.
+
+    Returns stdout (or None if outp is False).
+    '''
+    global auxverb
 
     if downp:
-        perhaps_down = downs['auxverb']
+        real_argv = auxverb + argv
     else:
-        perhaps_down = []
-
+        real_argv = argv
     if outp:
         stdout = subprocess.PIPE
     else:
         stdout = None
 
-    cmd = cmdl + cmd_list
-    if len(perhaps_down):
-        cmd = perhaps_down + cmd
-
-    (status, out) = execute_raw(cmdl[0], None, timeout,
-                                cmd, stdout=stdout)
+    (status, out, err) = execute_timeout(None, timeout, real_argv,
+                                         stdout=stdout)
 
     if status:
         bomb("%s%s failed (exit status %d)" %
-            ((downp and "(down) " or ""), cmdl[0], status))
+             ((downp and "(down) " or ""), argv, status))
+    if err:
+        bomb("%s unexpectedly produced stderr output `%s'" %
+             (argv, err))
 
     if outp and out and out[-1] == '\n':
         out = out[:-1]
@@ -225,170 +225,114 @@ def get_unix_socket(path):
     return s
 
 
-def expect(sock, search_str, timeout_sec, description=None):
-    debug('expect: "%s"' % search_str)
-    what = '"%s"' % (description or search_str or 'data')
+def expect(sock, search_bytes, timeout_sec, description=None, echo=False):
+    adtlog.debug('expect: "%s"' % search_bytes.decode())
+    what = '"%s"' % (description or search_bytes or 'data')
     out = b''
     with timeout(timeout_sec,
                  description and ('timed out waiting for %s' % what) or None):
         while True:
-            time.sleep(0.1)
             block = sock.recv(4096)
-            #debug('expect: got block: %s' % block)
+            if not block:
+                time.sleep(0.1)
+                continue
+            if echo:
+                sys.stderr.buffer.write(block)
             out += block
-            if search_str is None or search_str in out:
-                debug('expect: found "%s"' % what)
+            if search_bytes is None or search_bytes in out:
+                adtlog.debug('expect: found "%s"' % what)
                 break
 
+    return out
+
 
 def cmd_open(c, ce):
-    global downtmp
+    global auxverb, downtmp, downtmp_open
     cmdnumargs(c, ce)
     if downtmp:
         bomb("`open' when already open")
     caller.hook_open()
-    opened1()
-    downtmp = caller.hook_downtmp()
-    return opened2()
-
-
-def downtmp_mktemp():
-    d = execute('mktemp -d /tmp/adt-run.XXXXXX', [], downp=True, outp=True)
-    execute('chmod 1777', [d], downp=True)
-    return d
-
-
-def downtmp_remove():
-    global downtmp
-    execute('rm -rf --', [downtmp], downp=True)
-
-perl_quote_re = re.compile('[^-+=_.,;:() 0-9a-zA-Z]')
-
-
-def perl_quote_1chargroup(m):
-    return '\\x%02x' % ord(m.group(0))
-
-
-def perl_quote(s):
-    return '"' + perl_quote_re.sub(perl_quote_1chargroup, s) + '"'
-
-
-def opened1():
-    global down, downkind, downs
-    debug("downkind = %s, down = %s" % (downkind, str(down)))
-    if downkind == 'auxverb':
-        downs = {'auxverb': down,
-                 'shstring': down + ['sh', '-c']}
-    elif downkind == 'shstring':
-        downs = {'shstring': down,
-                 'auxverb': ['perl', '-e', '''
-                @cmd=(''' + (','.join(map(perl_quote, down))) + ''');
-                my $shstring = pop @ARGV;
-                s/'/'\\\\''/g foreach @ARGV;
-                push @cmd, "'$_'" foreach @ARGV;
-                my $argv0=$cmd[0];
-                exec $argv0 @cmd;
-                die "$argv0: $!"''']}
-    debug("downs = %s" % str(downs))
-
-
-def opened2():
-    global downtmp, downs
-    debug("downtmp = %s" % (downtmp))
+    adtlog.debug("auxverb = %s, downtmp = %s" % (str(auxverb), downtmp))
+    downtmp = caller.hook_downtmp(downtmp_open)
+    if downtmp_open and downtmp_open != downtmp:
+        bomb('virt-runner failed to restore downtmp path %s, gave %s instead'
+             % (downtmp_open, downtmp))
+    downtmp_open = downtmp
     return [downtmp]
 
 
+def downtmp_mktemp(path):
+    '''Generate a downtmp
+
+    When a path is given, this is the downtmp that we created when opening the
+    testbed the first time. We always want to keep the same path between
+    resets, as built package trees sometimes refer to absolute paths and thus
+    fail if they get moved around.
+    '''
+    if path:
+        check_exec(['mkdir', '--mode=1777', '--parents', downtmp_open],
+                   downp=True)
+        return path
+    else:
+        d = check_exec(['mktemp', '--directory', '/tmp/adt-run.XXXXXX'],
+                       downp=True, outp=True)
+        check_exec(['chmod', '1777', d], downp=True)
+        return d
+
+
+def downtmp_remove():
+    global downtmp, auxverb
+    if downtmp:
+        execute_timeout(None, copy_timeout,
+                        auxverb + ['rm', '-rf', '--', downtmp])
+        downtmp = None
+
+
 def cmd_revert(c, ce):
-    global downtmp
+    global auxverb, downtmp, downtmp_open
     cmdnumargs(c, ce)
     if not downtmp:
         bomb("`revert' when not open")
-    if not 'revert' in caller.hook_capabilities():
+    if 'revert' not in caller.hook_capabilities():
         bomb("`revert' when `revert' not advertised")
     caller.hook_revert()
-    opened1()
-    downtmp = caller.hook_downtmp()
-    return opened2()
+    downtmp = caller.hook_downtmp(downtmp_open)
+    if downtmp_open and downtmp_open != downtmp:
+        bomb('virt-runner failed to restore downtmp path %s, gave %s instead'
+             % (downtmp_open, downtmp))
+    adtlog.debug("auxverb = %s, downtmp = %s" % (str(auxverb), downtmp))
+
+    return [downtmp]
 
 
-def cmd_execute(c, ce):
-    cmdnumargs(c, ce, 5, None)
+def cmd_reboot(c, ce):
+    global downtmp
+    cmdnumargs(c, ce)
     if not downtmp:
-        bomb("`execute' when not open")
-    debug_re = re.compile('debug=(\d+)\-(\d+)$')
-    debug_g = None
-    timeout = 0
-    envs = []
-    for kw in ce[6:]:
-        if kw.startswith('debug='):
-            if debug_g:
-                bomb("multiple debug= in execute")
-            m = debug_re.match(kw)
-            if not m:
-                bomb("invalid execute debug arg `%s'" % kw)
-            debug_g = m.groups()
-        elif kw.startswith('timeout='):
-            try:
-                timeout = int(kw[8:], 0)
-            except ValueError:
-                bomb("invalid timeout arg `%s'" % kw)
-        elif kw.startswith('env='):
-            es = kw[4:]
-            eq = es.find('=')
-            if eq <= 0:
-                bomb("invalid env arg `%s'" % kw)
-            envs.append((es[:eq], es[eq + 1:]))
-        else:
-            bomb("invalid execute kw arg `%s'" % kw)
+        bomb("`reboot' when not open")
+    if 'reboot' not in caller.hook_capabilities():
+        bomb("`reboot' when `reboot' not advertised")
 
-    rune = 'set -e; exec '
+    # save current downtmp; try a few locations, as /var/cache might be r/o
+    # (argh Ubuntu touch)
+    directories = '/var/cache /home'
+    check_exec(['sh', '-ec', 'for d in %s; do if [ -w $d ]; then '
+                'tar --create --absolute-names -f $d/autopkgtest-tmpdir.tar'
+                ''' '%s'; exit 0; fi; done; exit 1''' %
+                (directories, downtmp)],
+               downp=True, timeout=copy_timeout)
+    adtlog.debug('cmd_reboot: saved current downtmp, rebooting')
 
-    stdout = None
-    tfd = None
-    if debug_g:
-        rune += " 3>&1"
+    caller.hook_reboot()
 
-    for ioe in range(3):
-        rune += " %d%s%s" % (ioe, '<>'[ioe > 0],
-                             pipes.quote(ce[ioe + 2]))
-    if debug_g:
-        (tfd, hfd) = m.groups()
-        tfd = int(tfd)
-        rune += " %d>&3 3>&-" % tfd
-        stdout = int(hfd)
-
-    rune += '; '
-
-    rune += 'cd %s; ' % pipes.quote(ce[5])
-
-    for e in envs:
-        (en, ev) = map(urllib.unquote, e)
-        rune += "%s=%s " % (en, pipes.quote(ev))
-
-    cmdl = map(urllib.unquote, ce[1].split(','))
-    rune += 'exec ' + ' '.join(map(pipes.quote, cmdl))
-
-    cmdl = downs['shstring'] + [rune]
-
-    stdout_copy = None
-    try:
-        if isinstance(stdout, int):
-            stdout_copy = os.dup(stdout)
-        try:
-            (status, out) = execute_raw('target-cmd', None,
-                                        timeout, cmdl, stdout=stdout_copy,
-                                        stdin=devnull_read,
-                                        stderr=subprocess.PIPE)
-        except Timeout:
-            raise FailedCmd(['timeout'])
-    finally:
-        if stdout_copy is not None:
-            os.close(stdout_copy)
-
-    if out:
-        bomb("target command unexpected produced stdout"
-             " visible to us `%s'" % out)
-    return [str(status)]
+    # restore downtmp
+    check_exec(['sh', '-ec', 'for d in %s; do '
+                'if [ -e $d/autopkgtest-tmpdir.tar ]; then '
+                ' tar --extract --absolute-names -f $d/autopkgtest-tmpdir.tar;'
+                ' rm $d/autopkgtest-tmpdir.tar; exit 0; '
+                'fi; done; exit 1' % directories],
+               downp=True, timeout=copy_timeout)
+    adtlog.debug('cmd_reboot: restored downtmp after reboot')
 
 
 def get_downtmp_host():
@@ -409,12 +353,17 @@ def copytree(src, dst):
 
     for f in os.listdir(src):
         fsrc = os.path.join(src, f)
-        subprocess.check_call(['cp', '-at', dst, fsrc])
+        subprocess.check_call(['cp', '-r', '--preserve=timestamps,links',
+                               '--target-directory', dst, fsrc])
 
 
 def copyup_shareddir(tb, host, is_dir, downtmp_host):
-    debug('copyup_shareddir: tb %s, host %s, is_dir %s, downtmp_host %s' % (
-        tb, host, is_dir, downtmp_host))
+    adtlog.debug('copyup_shareddir: tb %s host %s is_dir %s downtmp_host %s'
+                 % (tb, host, is_dir, downtmp_host))
+
+    host = os.path.normpath(host)
+    tb = os.path.normpath(tb)
+    downtmp_host = os.path.normpath(downtmp_host)
 
     timeout_start(copy_timeout)
     try:
@@ -424,36 +373,33 @@ def copyup_shareddir(tb, host, is_dir, downtmp_host):
             tb = downtmp_host + tb[len(downtmp):]
         else:
             tb_tmp = os.path.join(downtmp, os.path.basename(host))
-            debug('copyup_shareddir: tb path %s is not already in downtmp, '
-                  'copying to %s' % (tb, tb_tmp))
-            cp = subprocess.Popen(downs['auxverb'] + ['cp', '-a', tb, tb_tmp],
-                                  preexec_fn=preexecfn)
-            cp.communicate()
-            if cp.returncode != 0:
-                bomb('copyup_shareddir: cp -a exited with code %i' %
-                     cp.returncode)
+            adtlog.debug('copyup_shareddir: tb path %s is not already in '
+                         'downtmp, copying to %s' % (tb, tb_tmp))
+            check_exec(['cp', '-r', '--preserve=timestamps,links', tb, tb_tmp],
+                       downp=True)
             # translate into host path
             tb = os.path.join(downtmp_host, os.path.basename(host))
 
         if tb == host:
             tb_tmp = None
         else:
-            debug('copyup_shareddir: tb(host) %s is not already at '
-                  'destination %s, copying' % (tb, host))
+            adtlog.debug('copyup_shareddir: tb(host) %s is not already at '
+                         'destination %s, copying' % (tb, host))
             if is_dir:
                 copytree(tb, host)
             else:
                 shutil.copy(tb, host)
 
         if tb_tmp:
-            (is_dir and shutil.rmtree or os.unlink)(tb_tmp)
+            adtlog.debug('copyup_shareddir: rm intermediate copy: %s' % tb)
+            check_exec(['rm', '-rf', tb_tmp], downp=True)
     finally:
         timeout_stop()
 
 
 def copydown_shareddir(host, tb, is_dir, downtmp_host):
-    debug('copydown_shareddir: host %s, tb %s, is_dir %s, downtmp_host %s' % (
-        host, tb, is_dir, downtmp_host))
+    adtlog.debug('copydown_shareddir: host %s tb %s is_dir %s downtmp_host %s'
+                 % (host, tb, is_dir, downtmp_host))
 
     host = os.path.normpath(host)
     tb = os.path.normpath(tb)
@@ -468,7 +414,21 @@ def copydown_shareddir(host, tb, is_dir, downtmp_host):
         else:
             host_tmp = os.path.join(downtmp_host, os.path.basename(tb))
             if is_dir:
-                shutil.rmtree(host_tmp, ignore_errors=True)
+                if os.path.exists(host_tmp):
+                    try:
+                        shutil.rmtree(host_tmp)
+                    except OSError as e:
+                        adtlog.warning('cannot remove old %s, moving it '
+                                       'instead: %s' % (host_tmp, e))
+                        # some undeletable files? hm, move it aside instead
+                        counter = 0
+                        while True:
+                            p = host_tmp + '.old%i' % counter
+                            if not os.path.exists(p):
+                                os.rename(host_tmp, p)
+                                break
+                            counter += 1
+
                 shutil.copytree(host, host_tmp, symlinks=True)
             else:
                 shutil.copy(host, host_tmp)
@@ -478,15 +438,9 @@ def copydown_shareddir(host, tb, is_dir, downtmp_host):
         if host == tb:
             host_tmp = None
         else:
-            subprocess.call(downs['auxverb'] + ['rm', '-rf', tb],
-                            preexec_fn=preexecfn)
-            cp = subprocess.Popen(downs['auxverb'] + ['cp', '-a', host, tb],
-                                  preexec_fn=preexecfn)
-            cp.communicate()
-            if cp.returncode != 0:
-                bomb('copydown_shareddir: cp -a exited with code %i' %
-                     cp.returncode)
-
+            check_exec(['rm', '-rf', tb], downp=True)
+            check_exec(['cp', '-r', '--preserve=timestamps,links', host, tb],
+                       downp=True)
         if host_tmp:
             (is_dir and shutil.rmtree or os.unlink)(host_tmp)
     finally:
@@ -495,14 +449,18 @@ def copydown_shareddir(host, tb, is_dir, downtmp_host):
 
 def copyupdown(c, ce, upp):
     cmdnumargs(c, ce, 2)
+    copyupdown_internal(ce[0], c[1:], upp)
+
+
+def copyupdown_internal(wh, sd, upp):
+    '''Copy up/down a file or dir.
+
+    wh: 'copyup' or 'copydown'
+    sd: (source, destination) paths
+    upp: True for copyup, False for copydown
+    '''
     if not downtmp:
-        bomb("`copyup'/`copydown' when not open")
-    isrc = 0
-    idst = 1
-    ilocal = 0 + upp
-    iremote = 1 - upp
-    wh = ce[0]
-    sd = c[1:]
+        bomb("%s when not open" % wh)
     if not sd[0] or not sd[1]:
         bomb("%s paths must be nonempty" % wh)
     dirsp = sd[0][-1] == '/'
@@ -514,35 +472,43 @@ def copyupdown(c, ce, upp):
     # most cases, it's testbed end is already in the downtmp dir
     downtmp_host = get_downtmp_host()
     if downtmp_host:
-        if upp:
-            copyup_shareddir(sd[0], sd[1], dirsp, downtmp_host)
-        else:
-            copydown_shareddir(sd[0], sd[1], dirsp, downtmp_host)
+        try:
+            if upp:
+                copyup_shareddir(sd[0], sd[1], dirsp, downtmp_host)
+            else:
+                copydown_shareddir(sd[0], sd[1], dirsp, downtmp_host)
+        except Timeout:
+            raise FailedCmd(['timeout'])
         return
 
+    isrc = 0
+    idst = 1
+    ilocal = 0 + upp
+    iremote = 1 - upp
+
     deststdout = devnull_read
     srcstdin = devnull_read
     remfileq = pipes.quote(sd[iremote])
     if not dirsp:
         rune = 'cat %s%s' % ('><'[upp], remfileq)
         if upp:
-            deststdout = file(sd[idst], 'w')
+            deststdout = open(sd[idst], 'w')
         else:
-            srcstdin = file(sd[isrc], 'r')
+            srcstdin = open(sd[isrc], 'r')
             status = os.fstat(srcstdin.fileno())
-            if status.st_mode & 0111:
+            if status.st_mode & 0o111:
                 rune += '; chmod +x -- %s' % (remfileq)
         localcmdl = ['cat']
     else:
         taropts = [None, None]
         taropts[isrc] = '-c .'
-        taropts[idst] = '-p -x --no-same-owner'
+        taropts[idst] = '--preserve-permissions --extract --no-same-owner'
 
         rune = 'cd %s; tar %s -f -' % (remfileq, taropts[iremote])
         if upp:
             try:
                 os.mkdir(sd[ilocal])
-            except (IOError, OSError), oe:
+            except (IOError, OSError) as oe:
                 if oe.errno != errno.EEXIST:
                     raise
         else:
@@ -550,40 +516,45 @@ def copyupdown(c, ce, upp):
                 remfileq, remfileq)
             ) + rune
 
-        localcmdl = ['tar', '-C', sd[ilocal]] + (
+        localcmdl = ['tar', '--directory', sd[ilocal]] + (
             ('%s -f -' % taropts[ilocal]).split()
         )
-    rune = 'set -e; ' + rune
-    downcmdl = downs['shstring'] + [rune]
+    downcmdl = auxverb + ['sh', '-ec', rune]
 
     if upp:
         cmdls = (downcmdl, localcmdl)
     else:
         cmdls = (localcmdl, downcmdl)
 
-    debug(str(["cmdls", str(cmdls)]))
-    debug(str(["srcstdin", str(srcstdin), "deststdout",
-          str(deststdout), "devnull_read", devnull_read]))
+    adtlog.debug(str(["cmdls", str(cmdls)]))
+    adtlog.debug(str(["srcstdin", str(srcstdin), "deststdout",
+                      str(deststdout), "devnull_read", devnull_read]))
 
     subprocs = [None, None]
-    debug(" +< %s" % string.join(cmdls[0]))
+    adtlog.debug(" +< %s" % ' '.join(cmdls[0]))
     subprocs[0] = subprocess.Popen(cmdls[0], stdin=srcstdin,
                                    stdout=subprocess.PIPE,
                                    preexec_fn=preexecfn)
-    debug(" +> %s" % string.join(cmdls[1]))
+    adtlog.debug(" +> %s" % ' '.join(cmdls[1]))
     subprocs[1] = subprocess.Popen(cmdls[1], stdin=subprocs[0].stdout,
                                    stdout=deststdout,
                                    preexec_fn=preexecfn)
     subprocs[0].stdout.close()
-    timeout_start(copy_timeout)
-    for sdn in [1, 0]:
-        debug(" +" + "<>"[sdn] + "?")
-        status = subprocs[sdn].wait()
-        if not (status == 0 or (sdn == 0 and status == -13)):
-            timeout_stop()
-            bomb("%s %s failed, status %d" %
-                (wh, ['source', 'destination'][sdn], status))
-    timeout_stop()
+    try:
+        timeout_start(copy_timeout)
+        for sdn in [1, 0]:
+            adtlog.debug(" +" + "<>"[sdn] + "?")
+            status = subprocs[sdn].wait()
+            if not (status == 0 or (sdn == 0 and status == -13)):
+                timeout_stop()
+                bomb("%s %s failed, status %d" %
+                     (wh, ['source', 'destination'][sdn], status))
+        timeout_stop()
+    except Timeout:
+        for sdn in [1, 0]:
+            subprocs[sdn].kill()
+            subprocs[sdn].wait()
+        raise FailedCmd(['timeout'])
 
 
 def cmd_copydown(c, ce):
@@ -595,20 +566,36 @@ def cmd_copyup(c, ce):
 
 
 def cmd_shell(c, ce):
-    cmdnumargs(c, ce, 4)
+    cmdnumargs(c, ce, 1, None)
     if not downtmp:
         bomb("`shell' when not open")
+    # runners can provide a hook if they need a special treatment
     try:
-        caller.hook_shell(c[1], c[2], c[3], c[4])
+        caller.hook_shell(*c[1:])
     except AttributeError:
-        raise FailedCmd(['not supported by virt server'])
+        adtlog.debug('cmd_shell: using default shell command, dir %s' % c[1])
+        cmd = 'cd "%s"; ' % c[1]
+        for e in c[2:]:
+            cmd += 'export "%s"; ' % e
+        cmd += 'bash -i'
+        with open('/dev/tty', 'rb') as sin:
+            with open('/dev/tty', 'wb') as sout:
+                with open('/dev/tty', 'wb') as serr:
+                    subprocess.call(auxverb + ['sh', '-c', cmd],
+                                    stdin=sin, stdout=sout, stderr=serr)
 
 
 def command():
     sys.stdout.flush()
     while True:
         try:
-            ce = sys.stdin.readline()
+            ce = sys.stdin.readline().strip()
+            # FIXME: This usually means EOF (as checked below), but with Python
+            # 3 we often get empty strings here even though this is supposed to
+            # block for new input.
+            if ce == '':
+                time.sleep(0.1)
+                continue
             break
         except IOError as e:
             if e.errno == errno.EAGAIN:
@@ -619,10 +606,10 @@ def command():
     if not ce:
         bomb('end of file - caller quit?')
     ce = ce.rstrip().split()
-    c = map(urllib.unquote, ce)
+    c = list(map(url_unquote, ce))
     if not c:
         bomb('empty commands are not permitted')
-    debug('executing ' + string.join(ce))
+    adtlog.debug('executing ' + ' '.join(ce))
     c_lookup = c[0].replace('-', '_')
     try:
         f = globals()['cmd_' + c_lookup]
@@ -633,9 +620,9 @@ def command():
         if not r:
             r = []
         r.insert(0, 'ok')
-    except FailedCmd, fc:
+    except FailedCmd as fc:
         r = fc.e
-    print string.join(r)
+    print(' '.join(r))
 
 signal_list = [	signal.SIGHUP, signal.SIGTERM,
                 signal.SIGINT, signal.SIGPIPE]
@@ -648,13 +635,15 @@ def sethandlers(f):
 
 def cleanup():
     global downtmp, cleaning
-    debug("cleanup...")
+    adtlog.debug("cleanup...")
     sethandlers(signal.SIG_DFL)
-    cleaning = True
-    if downtmp:
-        caller.hook_cleanup()
-    cleaning = False
-    downtmp = False
+    # avoid recursion if something bomb()s in hook_cleanup()
+    if not cleaning:
+        cleaning = True
+        if downtmp:
+            caller.hook_cleanup()
+        cleaning = False
+        downtmp = None
 
 
 def error_cleanup():
@@ -663,23 +652,20 @@ def error_cleanup():
         try:
             cleanup()
             ok = True
-        except Quit, q:
-            print >> sys.stderr, q.m
+        except Quit as q:
+            sys.stderr.write(q.m)
+            sys.stderr.write('\n')
         except:
-            print >> sys.stderr, "Unexpected cleanup error:"
+            sys.stderr.write('Unexpected cleanup error:\n')
             traceback.print_exc()
-            print >> sys.stderr, ''
+            sys.stderr.write('\n')
         if not ok:
-            print >> sys.stderr, ("while cleaning up"
-                                  " because of another error:")
+            sys.stderr.write('while cleaning up because of another error:\n')
     except:
         pass
 
 
 def prepare():
-    global downtmp
-    downtmp = None
-
     def handler(sig, *any):
         cleanup()
         os.kill(os.getpid(), sig)
@@ -693,14 +679,15 @@ def mainloop():
     try:
         while True:
             command()
-    except Quit, q:
+    except Quit as q:
         error_cleanup()
         if q.m:
-            print >> sys.stderr, q.m
+            sys.stderr.write(q.m)
+            sys.stderr.write('\n')
         sys.exit(q.ec)
     except:
         error_cleanup()
-        print >> sys.stderr, "Unexpected error:"
+        sys.stderr.write('Unexpected error:\n')
         traceback.print_exc()
         sys.exit(16)
     finally:
